#!/usr/bin/perl -w

use Cowgen::Bunch;
use Cowgen::Dromed;
use Cowgen::Texture;
#use Cowgen::TextureList;
use Data::Dumper;
use MetaBrush::Dummy;

$sideNum=undef;
$brKind=undef;

$sides=undef;
$faces=undef;

$typenames = {
  '0' => 'solid()',
  '1' => 'air()',
  '2' => 'water()',
  '3' => 'flood()',
  '4' => 'evaporate()',
  '5' => 'solid2water()',
  '6' => 'solid2air()',
  '7' => 'air2solid()',
  '8' => 'water2solid()'
};

$kindnames = {
  '0' => 'cube()',
  '1' => 'cylinder()',
  '2' => 'pyramid()',
  '3' => 'apexpyramid()',
  '4' => 'wedge()',
  '5' => 'dodecahedron()'
};

$cubesidenames = {
  '0' => 'north()',
  '1' => 'west()',
  '2' => 'south()',
  '3' => 'east()',
  '4' => 'top()',
  '5' => 'bottom()'
};

$wedgesidenames = {
  '0' => 'side1()',
  '1' => 'side2()',
  '2' => 'side3()',
  '3' => 'top()',
  '4' => 'bottom()',
};

$sidenames = {
  '0' => 'side1()',
  '1' => 'side2()',
  '2' => 'side3()',
  '3' => 'side4()',
  '4' => 'side5()',
  '5' => 'side6()',
  '6' => 'side7()',
  '7' => 'side8()'
};

$sided = {
  '3' => 'threesided()',
  '4' => 'foursided()',
  '5' => 'fivesided()',
  '6' => 'sixsided()',
  '7' => 'sevensided()',
  '8' => 'eightsided()',
  '9' => 'ninesided()',
  '10' => 'tensided()'
};

$ltypenames = {
  '0' => 'omni()',
  '1' => 'spot()',
};

sub sideName {
  if ($brKind == cube()) {
    return $cubesidenames->{$sideNum};
  } elsif ($brKind == wedge()) {
    return $wedgesidenames->{$sideNum};
  } else {
    if ($sideNum == ($faces - 1)) {
      return 'bottom()';
    }
    if ($brKind != pyramid() && $brKind != pyramid()) {
      if ($sideNum == ($faces - 2)) {
        return 'top()';
      }
    }
    if (defined($sidenames->{$sideNum})) {
      return $sidenames->{$sideNum};
    } else {
      return $sideNum+1;
    }
  }
}

#my $gf = Cowgen::Gamfile->create();
my $gf = MetaBrush::Dummy->create(); # dummy for Cowgen::GamFile


# get commandline parameters
$infile = shift || die "Invalid input file\n";
$outfile = shift || 'out';
$templatefile = $infile;

# read lists from .mis
my $vbr = Cowgen::Vbr->create($infile);
my $bunch = Cowgen::Bunch->create('test');
my $textures = (Cowgen::Texture->create($vbr->getTXT()))->{'textnum'};

$bunch->fromBin($vbr,$gf);
my $blist=$bunch->{'blist'};
my $llist=$bunch->{'llist'};

# TEST
my $olist=$bunch->{'olist'};
my $rlist=$bunch->{'rlist'};

# open output file and add header
open FILE, ">$outfile" or die "Can't open file\n";

print FILE "\n#\n# Brushes\n#\n";

foreach $brush (@$blist) {

  my $type=undef;
  my $kind=undef;
  $sides=undef;
  $faces=undef;
  my $xyz=undef;
  my $wlh=undef;
  my $texture=undef;
  my $rotation=undef;
  my $tmp1;
  my $tmp2;
  my $tmp3;

  my $params=undef;
  my $i = 0;
  my $numSides;

  # Get brush type. What about flags? Not yet implemented.
  if (defined($typenames->{$brush->{'type'}})) {
    $type = $typenames->{$brush->{'type'}};
  } else {
    print "Unrecognized brush type $brush->{'type'}, including value in numeric form\n";
    $type = $brush->{'type'};
  }
  if (defined($brush->{'flags'})) {
    # Since the whole script is being built here,
    # assume the bunches default alignment is ALIGN_BY_VERTICES
    # and only add flags if this brush uses ALIGN_BY_SIDES
    my $flags=$brush->{'flags'};
    if ($flags != ALIGN_BY_VERTICES) {
      $type = "$type+$flags"; # Are there other flags? Include literal flags just in case
    }
  }

  # Get brush kind. Shape + sides?
  $brKind = $brush->{'kind'};
  if (defined($kindnames->{$brush->{'kind'}})) {
    $kind = $kindnames->{$brush->{'kind'}};
  } else {
    print "Unrecognized kind $brush->{'kind'}, including value in numeric form\n";
    $kind = $brush->{'kind'};
  }

  $faces = $#{$brush->{'faces'}}+1;
  $sides = $faces;
  if (($brKind != cube()) && ($brKind != wedge())) {
    if (($brKind == pyramid()) || ($brKind == apexpyramid())) {
      $sides--;
    }
    if ($brKind == cylinder()) {
      $sides-=2;
    }
    if (defined($sided->{$sides})) {
      $kind = "$kind+$sided->{$sides}";
    } else {
      print "Unrecognized sides $sides, including value in numeric form\n";
      $kind = "$kind+$sides}";
    }
  }

  # get xyz
  $xyz="$brush->{'x'},$brush->{'y'},$brush->{'z'}";

  # get wlh
  $wlh="$brush->{'w'},$brush->{'l'},$brush->{'h'}";

  # get texture
  $texture=$brush->{'texture'};
  if ($texture == deftex()) {
    $texture = 'deftex()';
  } elsif ($texture == jorge()) {
    $texture = 'jorge()';
  } elsif ($texture == sky()) {
    $texture = 'sky()';
  } elsif (defined($textures->{$texture})) {
    my ($j1,$j2,$j3,$j4,$j5,$tmp) = split ';', $textures->{$texture};
    $texture = "\$tl->texture('$tmp')";
  }

  # get rotation
  $rotation="$brush->{'rx'},$brush->{'ry'},$brush->{'rz'}";

  # put it together
  $params = "$type, $kind, $xyz, $wlh, $texture";
  if ($rotation ne '0,0,0') {
    $params = "$params, $rotation";    
  }

  print FILE "\$brush=\$bunch->addBrush( $params );\n";

  # handle texture params for faces
  for ($sideNum=0; $sideNum<$faces; $sideNum++) {
    my $faceName=sideName();
    my $faceTex = ($brush->{'faces'}[$sideNum])->{'t'};
    if ($faceTex != deftex()) {      
      if ($faceTex == jorge()) {
        $faceTex = 'jorge()';
      } elsif ($faceTex == sky()) {
        $faceTex = 'sky()';
      } else {
        my ($j1,$j2,$j3,$j4,$j5,$tmp) = split ';', $textures->{$faceTex};
        $faceTex = "\$tl->texture('$tmp')";
      }
      print FILE "  \$bunch->setTexture( \$brush, $faceName, $faceTex );\n";
    }
    my $faceU = ($brush->{'faces'}[$sideNum])->{'u'};
    if ($faceU != 0) {
      print FILE "  \$bunch->setFaceU( \$brush, $faceName, $faceU );\n";
    }
    my $faceV = ($brush->{'faces'}[$sideNum])->{'v'};
    if ($faceV != 0) {
      print FILE "  \$bunch->setFaceV( \$brush, $faceName, $faceV );\n";
    }
    my $faceScale = ($brush->{'faces'}[$sideNum])->{'s'};
    if ($faceScale != 16) {
      print FILE "  \$bunch->setFaceScale( \$brush, $faceName, $faceScale );\n";
    }
    my $faceRot = ($brush->{'faces'}[$sideNum])->{'r'};
    if ($faceRot != 0) {
      # Why is rotation way too large? Does fromBin read it correctly?
      # Temporarily try to fix it.
      #print FILE "  \$bunch->setFaceRot( \$brush, $faceName, $faceRot );\n";
      print FILE "  \$bunch->setFaceRot( \$brush, $faceName, $faceRot );\n";
    }
    my $faceAlign = ($brush->{'faces'}[$sideNum])->{'a'};
    if ($faceAlign != 0) {
      print FILE "  \$bunch->setFaceAlign( \$brush, $faceName, $faceAlign );\n";
    }
  }
}

print FILE "\n#\n# Lights\n#\n";

foreach $light (@$llist) {

  my $type=undef;
  my $xyz=undef;
  my $rotation=undef;
  my $color=undef;
  my $hue=undef;
  my $sat=undef;
  my $brt=undef;
  my $params=undef;

  my $i = 0;

  # Get light type.
  if (defined($ltypenames->{$light->{'type'}})) {
    $type = $ltypenames->{$light->{'type'}};
  } else {
    print "Unrecognized light type $light->{'type'}, including value in numeric form\n";
    $type = $light->{'type'};
  }

  # get xyz
  $xyz="$light->{'x'},$light->{'y'},$light->{'z'}";

  # get rotation
  $rotation="$light->{'rx'},$light->{'ry'},$light->{'rz'}";

  # get color
  # NOTE: rgb values read as hsb
  $hue=$light->{'h'};
  $sat=$light->{'s'};
  $brt=$light->{'b'};
  $color="{ hue=>$hue, saturation=>$sat, bright=>$brt }";

  # put it together
  $params = "$type, $xyz, $color";
  if ($rotation ne '0.00,0.00,0.00') {
    $params = "$params, $rotation";    
  }

  print FILE "\$lighttime=\$bunch->addLight( $params );\n";
}

close FILE;

print "Done.\n";

# EOF

1;
